#define _GNU_SOURCE

#include <errno.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <time.h>
#include <arpa/inet.h>
#include <stdarg.h>

#include <Conn.h>

/* Global variables */
static unsigned short	debug = 20;

static FILE	*Logf = NULL;
static char	*log_file = "s.log";
static int	port = 9000;
static short	ipv4 = 1, ipv6 = 1;
static int	maxconn = 0;

static void s_accept_error(struct Conn *C)
{
	Log(0, "%s: Cannot accept a new connection with id %llu!\n",
		__FUNCTION__, Conn_getid(C));
}

static void s_accept(struct Conn *C)
{
	int err;
	char	*s;

	Log(4, "%s (A connection was accepted from [%s/%d]. Enqueue greeting...\n",
		__FUNCTION__, C->addr, C->port);

	s = "Hello!\r\n";
	err = Conn_enqueue(C, s, strlen(s));
	if (err == -1) {
		Log(4, "\tCannot enqueue [%s]!\n", Conn_strerror());
		Conn_close(C);
	}
}

static void s_close(struct Conn *C)
{
	Log(5, "%s (A connection will be closed [%s/%d] id=%lld)\n",
		__FUNCTION__, C->addr, C->port, Conn_getid(C));
}

static int s_data_cb(struct Conn *C, char *line)
{
	char *dump;
	char buf[128];

	if (debug >= 8) {
		dump = Conn_dump(line, strlen(line));
		Log(8, "%s: recv head=%d tail=%d bytes on slot %d: [%s]\n",
			__FUNCTION__, C->ibuf_head, C->ibuf_tail,
			C->slot, dump);
		free(dump);
	}

	if (strcasecmp(line, "quit") == 0)
		exit(0);

	snprintf(buf, sizeof(buf), "%s\n", line);
	Log(8, "Send back: %s", buf);
	if (Conn_enqueue(C, buf, strlen(buf)) == -1) {
		Log(8, "ERROR: Cannot enqueue all data!\n");
		/* We return -1 and the current line will not be consumed */
		return -1;
	}

	return 0;
}

static void s_data(struct Conn *C)
{
	Conn_for_every_line(C, s_data_cb);
}

static void s_error(struct Conn *C)
{
	Log(1, "%s id=%llu (%s)\n",
		__FUNCTION__, Conn_getid(C), Conn_strerror());
}

int main(void)
{
	struct Conn *I4 = NULL, *I6 = NULL;
	int ret;
	struct timeval start;

	Logf = fopen(log_file, "w");
	if (!Logf) {
		fprintf(stderr, "Cannot open log file [%s] [%s]\n",
			log_file, strerror(errno));
		return 1;
	}
	if (debug > 0)
		setlinebuf(Logf);

	/*daemon(0, 0);*/

	gettimeofday(&start, NULL);
	Log(0, "\nStarting at %ld...\n", start.tv_sec);
	Log(0, "\tPort=%d\n", port);
	Log(0, "\tLogFile=%s Debug=%d\n", log_file, debug);
	Log(0, "\tipv4=%s ipv6=%s\n", (ipv4 == 1) ? "on" : "off", (ipv6 == 1) ? "on" : "off");

	/* set library output debug */
	Conn_debug(Logf, debug);

	ret = Conn_init(maxconn);
	if (ret == -1) {
		Log(0, "%s", Conn_strerror());
		return 1;
	}

	if (ipv4 == 1) {
		Log(9, "Try to register IPv4 socket...\n");
		I4 = Conn_socket(PF_INET, SOCK_STREAM, port);
		if (!I4)
			Log(1, "Cannot bind on ipv4 socket [%s].\n", Conn_strerror());
	}

	if (ipv6 == 1) {
		Log(9, "Try to register IPv6 socket...\n");
		I6 = Conn_socket(PF_INET6, SOCK_STREAM, port);
		if (!I6)
			Log(1, "Cannot bind on ipv6 socket [%s].\n", Conn_strerror());
	}

	if ((I6 == NULL) && (I4 == NULL)) {
		Log(0, "Cannot bind!\n");
		return 1;
	}

	Conn_accept_cb = s_accept;
	Conn_data_cb = s_data;
	Conn_close_cb = s_close;
	Conn_error_cb = s_error;
	Conn_accept_error_cb = s_accept_error;

	while (1) {
		/* -1 we do not have any other business but wait for an event */
		ret = Conn_poll(-1);
		if (ret == -1) {
			Log(0, "Error calling Conn_poll [%s]!\n", Conn_strerror());
			return 1;
		} else if (ret == 0)
			break;
	}

	Conn_shutdown();

	Log(0, "Finish!\n\n");

	return 0;
}
